﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using CheesyFX;
using GPUTools.Physics.Scripts.Behaviours;
using Leap.Unity;
using MeshVR;
using MVR.FileManagementSecure;
using Oculus.Platform;
using SimpleJSON;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;

/***********************************************************************************************************************************
 * prestigitis HeelAdjust 1.0.2
 * 
 * prestigitis_HeelAdjust.cs 
 * 
 * requires VaM 1.19.2.2
 * add this script to a Person atom to adjust foot pose and add heel and toe colliders for high-heeled or thick-soled shoes. 
 * autoload is triggered by mouse left button or controller button release while the 'clothing' or 'clothing preset' tab is active.
 * for oculus touch, autoload is triggered only when using touch controller button A.
 * 
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 ************************************************************************************************************************************
 * 1.0.0 (20200109) prestigitis
 *     - original release
 * 1.0.1 
 *     JayJayWon
 *     - fixed for VaM 1.19 changes in VaM clothing hierarchy
 *     - fixed for VaM 1.19 changes in clothing item naming convention
 *     - saved and unsaved active clothing checks combined into one method
 *     prestigitis
 *     - suppress repeated errors in FixedUpdate when Person UI has not been loaded
 *     - automagically create HeelAdjust folder if it does not exist
 * 1.0.2
 *     ICannotDie
 *     - added support for male models. See SetModelGender().
 * 2.0
 *     CheesyFX
 *     - nearly complete rewrite
 *     - OnDestroy() fixed: you can safely disable the plugin, reload or remove it, save it with the scene and load it afterwards.
 *     - the current settings are stored with the scene, so the user doesn't necessarily has to have the preset you were using. 
 *     - more options while designing the colliders (scale, rotate and offset in any direction)
 *     - option to completely lock or just restrict the toes
 *     - option to auto-adjust controller heights (only feet or main controller)
 *     - presets are triggered when a clothing item with an existing preset is equipped by any means.
 *       If you e.g. trigger different outfits through buttons in your scene, the matching preset will automatically load with it.
 *     - presets created with the original plugin get converted and should be more or less compatible, see LegacyLoad(string path)
 *     - various other fixes and improvements to optimize performance, mainly creating references instead of repeatedly using Transform.Find()
 *     - UI-overhaul, using custom Vector3 sliders and a tab bar
 ***********************************************************************************************************************************/

namespace CheesyFX
{
    public class HeelAdjust2 : MVRScript
    {
        /***********************************************************************************************************************************/
        /* configuration settings 
        /***********************************************************************************************************************************/
        private string HeelAdjustSubfolderName = "HeelAdjust";           // name of subfolder in which heeladjust files are saved in Saves folder
        private const string saveDir = "Saves/PluginData/CheesyFX/HeelAdjust2";
        private string bundlePath = "Custom/Scripts/CheesyFX/HeelAdjust2/footsteps.audiobundle";
        private Color ShoeColliderColor = new Color(0.0f, 0f, 1f, 0.7f); // color of the toe collider when visible
        /***********************************************************************************************************************************/

        protected JSONStorableFloat footSlope = new JSONStorableFloat("Foot Slope", 0f, -40f, 80f, true);
        protected JSONStorableFloat footSlopeMax = new JSONStorableFloat("Max Foot Slope", 90f, 0f, 150f, true);
        private JSONStorableFloat lFootSlopeDelta = new JSONStorableFloat("Left Foot Slope Delta", 0f, -180f, 180f);
        private JSONStorableFloat rFootSlopeDelta = new JSONStorableFloat("Right Foot Slope Delta", 0f, -180f, 180f);
        private JSONStorableBool autoCalculateDeltas = new JSONStorableBool("Auto Calculate Deltas", true);
        private JSONStorableBool showColliders = new JSONStorableBool("Show Colliders", false);
        private string settingsOutputFormat;
        private UIDynamicTextField _infoTextField;
        private JSONStorableStringChooser itemToStore = new JSONStorableStringChooser("itemToStore", new List<string>{"None"}, "None", "Item To Store");
        private JSONStorableStringChooser heightAdjustMode = new JSONStorableStringChooser("adjustControllerHeight", new List<string>{"None", "Feet", "Root"}, "None", "Adjust Controller Heights");

        private float lDelta;
        private float rDelta;
        
        private Vector3 rToeBaseRotation = new Vector3(-14.4f, 0f, -1.8f);
        private Vector3 lToeBaseRotation = new Vector3(-14.4f, 0f, 1.8f);
        private Vector3 rHeelBaseRotation = new Vector3(-18.7f, 0f, -2.3f);
        private Vector3 lHeelBaseRotation = new Vector3(-18.7f, 0f, 2.3f);
        // private string gender = "Female";
        private List<string> existingPresets;
        private string currentPreset;
        private DAZClothingItem currentItem;
        private JSONClass sceneData;
        private int currentHeightAdjustMode;
        public static Action<DAZClothingItem, bool> onClothingItemToggled;

        private DAZBone rFoot;
        private DAZBone lFoot;
        private DAZBone rToe;
        private DAZBone lToe;
        private List<Transform> toes => lToes.Concat(rToes).ToList();
        private List<Transform> lToes = new List<Transform>();
        private List<Transform> rToes = new List<Transform>();
        private Transform clothes;
        private FreeControllerV3 root, rfootCTRL, lfootCTRL, rtoeCTRL, ltoeCTRL;
        private Vector3 rfootJointDrive, lfootJointDrive, ltoeJointDrive, rtoeJointDrive;
        private float rFootRotX, lFootRotX, rFootY, lFootY, rootY;
        private Transform rHeelColliderBase, rToeColliderBase, lHeelColliderBase, lToeColliderBase;
        private Transform rHeelColTOffset = new GameObject{name="rHeelColTOffset"}.transform;
        private Transform rHeelColROffset = new GameObject{name="rHeelColROffset"}.transform;
        private Transform lHeelColTOffset = new GameObject{name="lHeelColTOffset"}.transform;
        private Transform lHeelColROffset = new GameObject{name="lHeelColROffset"}.transform;

        private Transform rToeColTOffset = new GameObject{name="rToeColTOffset"}.transform;
        private Transform rToeColROffset = new GameObject{name="rToeColROffset"}.transform;
        private Transform lToeColTOffset = new GameObject{name="lToeColTOffset"}.transform;
        private Transform lToeColROffset = new GameObject{name="lToeColROffset"}.transform;
        
        public static BoxCollider rToeCollider;
        public static BoxCollider lToeCollider;
        public static BoxCollider rHeelCollider;
        public static BoxCollider lHeelCollider;

        private MyJSONStorableVector3 heelColliderScale =
            new MyJSONStorableVector3("HeelColScale", new Vector3(.02f, 0f, .02f), Vector3.zero, new Vector3(.1f,.4f,.2f));
        private MyJSONStorableVector3 heelColliderOffset =
            new MyJSONStorableVector3("HeelColOffset", new Vector3(0f, 0f, 0f), new Vector3(-.03f, -.1f, -.02f), new Vector3(.03f, .1f, .1f));
        private MyJSONStorableVector3 heelColliderRotation =
            new MyJSONStorableVector3("HeelColRot", new Vector3(0f, 0f, 0f), new Vector3(-20f,-45f,-20f), new Vector3(20f,45f,20f));
        private MyJSONStorableVector3 toeColliderScale =
            new MyJSONStorableVector3("ToeColScale", new Vector3(.05f, 0f, .05f), Vector3.zero, new Vector3(.1f, .4f,.1f));
        private MyJSONStorableVector3 toeColliderOffset =
            new MyJSONStorableVector3("ToeColOffset", new Vector3(0f, 0f, 0f), new Vector3(-.1f, -.1f, -.1f), new Vector3(.1f, .1f, .1f));
        private MyJSONStorableVector3 toeColliderRotation =
            new MyJSONStorableVector3("ToeColRot", new Vector3(0f, 0f, 0f), new Vector3(-20f,-20f,-20f), new Vector3(20f,20f,20f));
        private MyJSONStorableVector3 toeRotation =
            new MyJSONStorableVector3("ToeRotation", new Vector3(0f, 0f, 0f), new Vector3(-40f,-10f,-10f), new Vector3(40f,10f,10f));
        private MyJSONStorableVector3 toeLimits =
            new MyJSONStorableVector3("ToeLimits", new Vector3(0f, 0f, 0f), new Vector3(0f,0f,0f), new Vector3(20f,20f,20f));
        private JSONStorableBool lockToes = new JSONStorableBool("Lock Toe Joint", false);
        private JSONStorableBool lockIndividualToes = new JSONStorableBool("Lock Individual Toes", false);
        private JSONStorableBool collisionEnabled = new JSONStorableBool("Collision Enabled", true);
        
        private JSONStorableBool footStepsEnabled = new JSONStorableBool("Footsteps Enabled", true);
        private JSONStorableFloat spacialBlend = new JSONStorableFloat("Spacial Blend", 1f, 0f, 1f);
        private JSONStorableFloat baseVolume = new JSONStorableFloat("Base Volume", 5f, 0f, 50f, false);
        private JSONStorableFloat basePitch = new JSONStorableFloat("Base Pitch", 1f, 0f, 2f);
        private JSONStorableStringChooser soundLibrary = new JSONStorableStringChooser("soundLibrary", null, "heels", "Sound Library");
        private JSONStorableStringChooser soundProducers = new JSONStorableStringChooser("soundProducers",
            new List<string>{"Everything", "Floors, Walls, Glass, CUAs", "Floors, Walls, Glass", "CUAs", "Only this..."}, "", "Sound Producers");
        public static JSONStorableStringChooser onlySoundFromThis = new JSONStorableStringChooser("onlySoundFromThis",
            null, "", "Only this");

        private JSONArray linkedItemsJSON;

        private UIDynamicButton reloadButton;
        private UIDynamicButton toggleItemButton;

        // private UIDynamicV3Slider toeRotationUid;
        // private UIDynamicV3Slider toeLimitsUid;
        private UIDynamicV3Slider colScaleUid;
        private UIDynamicV3Slider colOffsetUid;
        private UIDynamicV3Slider colRotationUid;
        private UIDynamicSlider footSlopeUid;

        private CollisionHandler rToeCollisionHandler;
        private CollisionHandler lToeCollisionHandler;
        private CollisionHandler rHeelCollisionHandler;
        private CollisionHandler lHeelCollisionHandler;

        private List<object> UIElements = new List<object>();
        
        private void OnShowCollidersChanged(bool b)
        {
            if (!rToeCollider) return;
            rToeCollider.transform.GetComponent<Renderer>().enabled = b;
            rHeelCollider.transform.GetComponent<Renderer>().enabled = b;
            lToeCollider.transform.GetComponent<Renderer>().enabled = b;
            lHeelCollider.transform.GetComponent<Renderer>().enabled = b;
        }

        // private void SyncClothingChoices()
        // {
        //     itemToStore.choices = GetActiveClothes(true);
        //     if (!itemToStore.choices.Contains(itemToStore.val)) itemToStore.val = "None";
        // }
        
        void OnEnable()
        {
            if(rToeColliderBase != null)
            {
                // ZeroJointDrives();
                // lDelta = lFootSlopeDelta.val;
                // rDelta = rFootSlopeDelta.val;
                rToeColliderBase.gameObject.SetActive(true);
                rHeelColliderBase.gameObject.SetActive(true);
                lToeColliderBase.gameObject.SetActive(true);
                lHeelColliderBase.gameObject.SetActive(true);
                OnShowCollidersChanged(showColliders.val);
                if(currentPreset != null) LoadFromPath(currentPreset);
                else if (sceneData != null)
                {
                    LoadFromJSON(sceneData);
                    reloadButton.label = "Scene data loaded";
                }
                if(autoCalculateDeltas.val) CalculateDeltas();
                rToeCollisionHandler.enabled = lToeCollisionHandler.enabled =
                    rHeelCollisionHandler.enabled = lHeelCollisionHandler.enabled = footStepsEnabled.val;
            }
            
        }

        private ConfigurableJoint lFootCtrlLink;
        private ConfigurableJoint rFootCtrlLink;
        private ConfigurableJoint lFootShinLink;
        private ConfigurableJoint rFootShinLink;
        
        private ConfigurableJoint lToeFootLink;
        private ConfigurableJoint rToeFootLink;
        public override void Init()
        {
            try
            {
                FileManagerSecure.CreateDirectory(saveDir);
                string packagePath = this.GetPackagePath();
                if (packagePath != String.Empty) bundlePath = $"{packagePath}{bundlePath}";
                root = containingAtom.mainController;
                rfootCTRL = containingAtom.freeControllers.First(fc => fc.name == "rFootControl");
                rtoeCTRL = containingAtom.freeControllers.First(fc => fc.name == "rToeControl");
                lfootCTRL = containingAtom.freeControllers.First(fc => fc.name == "lFootControl");
                ltoeCTRL = containingAtom.freeControllers.First(fc => fc.name == "lToeControl");
                ltoeCTRL.currentRotationState = FreeControllerV3.RotationState.Off;
                rtoeCTRL.currentRotationState = FreeControllerV3.RotationState.Off;
                ltoeCTRL.currentPositionState = FreeControllerV3.PositionState.Off;
                rtoeCTRL.currentPositionState = FreeControllerV3.PositionState.Off;
                
                var dazBones = containingAtom.GetComponentsInChildren<DAZBone>();
                rFoot = dazBones.First(x => x.name == "rFoot");
                lFoot = dazBones.First(x => x.name == "lFoot");
                rToe = dazBones.First(x => x.name == "rToe");
                lToe = dazBones.First(x => x.name == "lToe");
                foreach (var j in lFoot.GetComponents<ConfigurableJoint>())
                {
                    if (j.connectedBody.GetComponent<FreeControllerV3>()) lFootCtrlLink = j;
                    else lFootShinLink = j;
                }
                foreach (var j in rFoot.GetComponents<ConfigurableJoint>())
                {
                    if (j.connectedBody.GetComponent<FreeControllerV3>()) rFootCtrlLink = j;
                    else rFootShinLink = j;
                }
                foreach (var j in lToe.GetComponents<ConfigurableJoint>())
                {
                    if (!j.connectedBody.GetComponent<FreeControllerV3>()) lToeFootLink = j;
                }
                foreach (var j in rToe.GetComponents<ConfigurableJoint>())
                {
                    if (!j.connectedBody.GetComponent<FreeControllerV3>()) rToeFootLink = j;
                }

                foreach (Transform c in rToe.transform)
                {
                    if (c.name.Contains("BigToe") || c.name.Contains("SmallToe")) rToes.Add(c);
                }
                foreach (Transform c in lToe.transform)
                {
                    if (c.name.Contains("BigToe") || c.name.Contains("SmallToe")) lToes.Add(c);
                }
                
                GetExistingPresets();
                var gender = containingAtom.GetComponentInChildren<DAZCharacterSelector>()?.gender == DAZCharacterSelector.Gender.Male ? "Male" : "Female";
                clothes = containingAtom.gameObject.transform.Find($"rescale2/geometry/{gender}Clothes");
                clothes.gameObject.AddComponent<ChildrenChangedEventHandler>();

                onClothingItemToggled = OnClothingItemToggled;

                itemToStore.setCallbackFunction += val =>
                {
                    if (existingPresets.Any(x => x.EndsWith($"{val}.json")))
                    {
                        saveButton.buttonColor = new Color(1f, 0.34f, 0.34f);
                        saveButton.label = "Override preset";
                    }
                    else
                    {
                        saveButton.buttonColor = new Color(0.49f, 1f, 0.49f);
                        saveButton.label = "Save new preset";
                    }
                };

                toeColliderScale.setCallbackFunction += OnToeScaleChanged;
                toeColliderOffset.setCallbackFunction += OnToeOffsetChanged;
                toeColliderRotation.setCallbackFunction += OnToeRotationChanged;

                heelColliderScale.setCallbackFunction += val => OnHeelScaleChanged(val);
                heelColliderOffset.setCallbackFunction += OnHeelOffsetChanged;
                heelColliderRotation.setCallbackFunction += OnHeelRotationChanged;

                toeRotation.setCallbackFunction += val => OnFootPoseChanged();

                lockToes.setCallbackFunction += ToeSetLocked;
                lockIndividualToes.setCallbackFunction += IndivToesSetLocked;

                collisionEnabled.setCallbackFunction += OnCollisionEnabledChanged;

                heightAdjustMode.setCallbackFunction += val => StartCoroutine(OnHeightAdjustModeChanged(val, true));
                
                footSlope.setCallbackFunction += val => OnFootPoseChanged();
                lFootSlopeDelta.setCallbackFunction += val =>
                {
                    lDelta = val;
                    OnFootPoseChanged();
                };
                rFootSlopeDelta.setCallbackFunction += val =>
                {
                    rDelta = val;
                    OnFootPoseChanged();
                };
                
                showColliders.setCallbackFunction += OnShowCollidersChanged;

                footStepsEnabled.setCallbackFunction += val =>
                {
                    lHeelCollisionHandler.enabled = rHeelCollisionHandler.enabled = val;
                    lToeCollisionHandler.enabled = rToeCollisionHandler.enabled = val;
                    val.Print();
                };
                baseVolume.setCallbackFunction += val => AudioManager.baseVolume = val;
                basePitch.setCallbackFunction += val => AudioManager.basePitch = val;
                spacialBlend.setCallbackFunction += val =>
                {
                    lHeelCollisionHandler.audioSource.spatialBlend =
                        rHeelCollisionHandler.audioSource.spatialBlend = val;
                    lToeCollisionHandler.audioSource.spatialBlend = 
                        rToeCollisionHandler.audioSource.spatialBlend = val;
                };
                soundLibrary.setCallbackFunction += val => GetClipLibraryFromAssetBundle(footstepBundle, val);
                soundProducers.setCallbackFunction += val => AudioManager.SetSoundProducers(val, false);
                onlySoundFromThis.setCallbackFunction += val => AudioManager.SetSoundProducers(val, true);

                
                
                mainTabbar = CreateTabBar(new[] { "Scene & Pose Setup", "Create Preset" }, SwitchTab);
                mainTabbar.SelectTab(1);
                mainTabbar.SelectTab(0);
                
                RegisterFloat(lFootSlopeDelta);
                RegisterFloat(rFootSlopeDelta);
                RegisterFloat(footSlopeMax);
                RegisterBool(collisionEnabled);
                RegisterBool(autoCalculateDeltas);
                RegisterBool(footStepsEnabled);
                RegisterFloat(baseVolume);
                RegisterFloat(basePitch);
                RegisterFloat(spacialBlend);
                RegisterStringChooser(soundLibrary);
                RegisterStringChooser(soundProducers);
                RegisterStringChooser(onlySoundFromThis);
                
                StartCoroutine(DeferredInit());

                onlySoundFromThis.choices = SuperController.singleton.GetAtoms().Select(x => x.uid).ToList();
                SuperController.singleton.onAtomUIDsChangedHandlers += SyncAtoms;
                
                var pmc = containingAtom.presetManagerControls.First(x => x.name == "PosePresets");
                pm = pmc.GetComponent<PresetManager>();
                pm.postLoadEvent.AddListener(CalculateDeltasFromJSON);
                // SuperController.singleton.GetAtoms().ForEach(x => x.type.Print());
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private PresetManager pm;
        private string reloadLabel = "<b>Reload</b>: None";
        private string toggleItemLabel = "<b>Toggle</b>: None";
        
        private IEnumerator DeferredInit()
        {
            while (SuperController.singleton.isLoading) yield return null;
            yield return new WaitForEndOfFrame();
            StorePoseData();
            AddToeColliders();
            AddHeelColliders();
            OnShowCollidersChanged(showColliders.val);
            HandShakeWithFlexIt();
            for (int i = 0; i < clothes.childCount; i++)
            {
                clothes.GetChild(i).gameObject.AddComponent<ClothingEquippedEventHandler>();
            }
            if (sceneData != null)
            {
                reloadButton.label = "Scene data loaded";
                LoadFromJSON(sceneData);
            }
            CalculateDeltas();
            soundProducers.val = "Floors, Walls, Glass, CUAs";
            LoadBundle();
        }

        private void OnDisable()
        {
            rToeColliderBase.gameObject.SetActive(false);
            rHeelColliderBase.gameObject.SetActive(false);
            lToeColliderBase.gameObject.SetActive(false);
            lHeelColliderBase.gameObject.SetActive(false);
            Reset();
            lockToes.val = false;
            lockIndividualToes.val = false;
            rToeCollisionHandler.enabled = lToeCollisionHandler.enabled =
                rHeelCollisionHandler.enabled = lHeelCollisionHandler.enabled = false;
        }

        private void OnDestroy()
        {
            OnDisable();
            GPUCollidersManager.DeregisterLineSphereCollider(lHeelGPUCapsule);
            GPUCollidersManager.DeregisterLineSphereCollider(lToeGPUCapsule);
            GPUCollidersManager.DeregisterLineSphereCollider(rHeelGPUCapsule);
            GPUCollidersManager.DeregisterLineSphereCollider(rToeGPUCapsule);
            Destroy(rToeColliderBase.gameObject);
            Destroy(rHeelColliderBase.gameObject);
            Destroy(lToeColliderBase.gameObject);
            Destroy(lHeelColliderBase.gameObject);
            
            for (int i = 0; i < clothes.childCount; i++)
            {
                Destroy(clothes.GetChild(i).GetComponent<ClothingEquippedEventHandler>());
            }
            Destroy(clothes.GetComponent<ChildrenChangedEventHandler>());
            HandShakeWithFlexIt();
            pm.postLoadEvent.RemoveListener(CalculateDeltasFromJSON);
            Destroy(rToeCollisionHandler);
            Destroy(lToeCollisionHandler);
            Destroy(rHeelCollisionHandler);
            Destroy(lHeelCollisionHandler);
            AssetLoader.DoneWithAssetBundleFromFile(footstepBundle.path);
            SuperController.singleton.onAtomUIDsChangedHandlers -= SyncAtoms;
        }
        
        private void SwitchTab(int id)
        {
            RemoveUIElements(UIElements);
            RemoveUIElements(UIElementsSliders);
            RemoveUIElements(UIElementsAdditionalItems);
            if(id == 0) CreateSceneUI();
            else CreatePresetUI();
        }

        private void CreateSceneUI()
        {
            // if(lockedInfo != null) RemoveTextField(lockedInfo);
            UIElements.Add(CreateSlider(lFootSlopeDelta));
            UIElements.Add(CreateSlider(rFootSlopeDelta, true));
            UIElements.Add(CreateSlider(footSlopeMax));
            var button = CreateButton("Recalculate Deltas");
            button.button.onClick.AddListener(() => StartCoroutine(IECalculateDeltas()));
            UIElements.Add(button);
            UIElements.Add(CreateToggle(autoCalculateDeltas));

            var popup = CreateScrollablePopup(heightAdjustMode, true);
            popup.popupPanelHeight = 220f;
            UIElements.Add(popup);
            
            toggleItemButton = CreateButton("", true);
            toggleItemButton.label = toggleItemLabel;
            toggleItemButton.GetComponentInChildren<Text>().fontSize = 24;
            toggleItemButton.button.onClick.AddListener(() => ToggleClothingItem(currentItem));
            UIElements.Add(toggleItemButton);
            UIElements.Add(CreateToggle(collisionEnabled,true));
            UIElements.Add(CreateToggle(showColliders, true));
            var textField = CreateTextField(new JSONStorableString("", "Foot Steps"));
            textField.height = 50f;
            UIElements.Add(textField);
            var spacer = CreateSpacer(true);
            spacer.height = 50f;
            UIElements.Add(spacer);
            UIElements.Add(CreateToggle(footStepsEnabled));
            UIElements.Add(CreateFilterablePopup(soundLibrary));
            UIElements.Add(CreateSlider(baseVolume));
            UIElements.Add(CreateSlider(basePitch));
            UIElements.Add(CreateSlider(spacialBlend));
            
            UIElements.Add(CreatePopup(soundProducers, true));
            UIElements.Add(CreateFilterablePopup(onlySoundFromThis, true));
        }

        private void CreatePresetUI()
        {
            footSlopeUid = CreateSlider(footSlope);
            footSlopeUid.label = lockToes.val ? "Foot Slope (unlock toe)" : "Foot Slope";
            UIElements.Add(footSlopeUid);
            UIElements.Add(CreateFilterablePopup(itemToStore, true));
            
            saveButton = CreateButton("Save/overwrite preset for above item", true);
            saveButton.button.onClick.AddListener(() => SavePreset(itemToStore.val));
            UIElements.Add(saveButton);

            reloadButton = CreateButton("", true);
            reloadButton.label = reloadLabel;
            reloadButton.GetComponentInChildren<Text>().fontSize = 24;
            reloadButton.button.onClick.AddListener(() => LoadFromPath(currentPreset));
            UIElements.Add(reloadButton);
            
            linkAdditionalButton = CreateButton("Link Additional Items", true);
            linkAdditionalButton.button.onClick.AddListener(CreateAdditionalItemsUI);
            UIElements.Add(linkAdditionalButton);
            SyncLinkAdditionalButton();
            
            UIElements.Add(CreateToggle(lockToes));
            UIElements.Add(CreateToggle(lockIndividualToes));
            UIElements.Add(CreateToggle(showColliders));
            if (!lockToes.val) CreateSliders();
            else
            {
                lockedInfo = CreateTextField(new JSONStorableString("", "Unlock the toe joint to edit the colliders."));
                lockedInfo.height = 100f;
                UIElements.Add(lockedInfo);
            }
        }

        private void CreateSliders()
        {
            RemoveUIElements(UIElementsAdditionalItems);
            if(lockedInfo != null) RemoveTextField(lockedInfo);
            var toeRotationUid = CreateV3Slider(toeRotation);
            var toeLimitsUid = CreateV3Slider(toeLimits);
            
            var tabbar = CreateTabBar(new[] { "Heel Collider", "Toe Collider" }, RegisterSliders);
            
            colScaleUid = CreateV3Slider();
            colOffsetUid = CreateV3Slider();
            colRotationUid = CreateV3Slider();
            colOffsetUid.sliders[1].gameObject.SetActive(false);
            
            UIElementsSliders.Add(toeRotationUid);
            UIElementsSliders.Add(toeLimitsUid);
            UIElementsSliders.Add(colScaleUid);
            UIElementsSliders.Add(colOffsetUid);
            UIElementsSliders.Add(colRotationUid);
            UIElementsSliders.Add(tabbar);
            
            tabbar.SelectLast();
        }

        private UIDynamicTabBar mainTabbar;
        private UIDynamicButton saveButton;
        private UIDynamicButton linkAdditionalButton;
        private UIDynamicTextField lockedInfo;
        private List<object> UIElementsSliders = new List<object>();
        private List<object> UIElementsAdditionalItems = new List<object>();
        private Dictionary<string, DAZClothingItem> activeItems = new Dictionary<string, DAZClothingItem>();
        private JSONStorableStringChooser
            aditionalItemChooser = new JSONStorableStringChooser("", null, "None", "Item to Link");

        private List<string> linkedItems = new List<string>();
        private JSONStorableString linkInfo = new JSONStorableString("linkInfo", "");

        private void CreateAdditionalItemsUI()
        {
            RemoveTextField(lockedInfo);
            RemoveUIElements(UIElementsSliders);
            linkAdditionalButton.label = "Return";
            linkAdditionalButton.button.onClick.RemoveAllListeners();
            linkAdditionalButton.button.onClick.AddListener(() => SwitchTab(1));
            UIElementsAdditionalItems.Add(linkAdditionalButton);
            UIElementsAdditionalItems.Add(CreateFilterablePopup(aditionalItemChooser));
            var button = CreateButton("Link", true);
            button.button.onClick.AddListener(() => LinkItem(aditionalItemChooser.val, true));
            UIElementsAdditionalItems.Add(button);
            button = CreateButton("Unlink", true);
            button.button.onClick.AddListener(() => LinkItem(aditionalItemChooser.val, false));
            UIElementsAdditionalItems.Add(button);
            button = CreateButton("Reset", true);
            button.button.onClick.AddListener(ResetLinkedItems);
            UIElementsAdditionalItems.Add(button);
            var textfield = CreateTextField(linkInfo);
            textfield.height = 600f;
            UIElementsAdditionalItems.Add(textfield);
        }

        private void SyncAtoms(List<string> uids)
        {
            onlySoundFromThis.choices = new List<string>(uids);
            if(soundProducers.val!= "Only this...") AudioManager.SetSoundProducers(soundProducers.val, false);
            else if (!uids.Contains(onlySoundFromThis.val)) onlySoundFromThis.val = null;
        }

        private void LinkItem(string itemName, bool val)
        {
            if (val) linkedItems.Add(itemName);
            else linkedItems.Remove(itemName);
            SyncLinkInfo();
        }

        private void ResetLinkedItems()
        {
            linkedItems.Clear();
            SyncLinkInfo();
        }

        private void SyncLinkInfo()
        {
            linkInfo.val = string.Join("\n", linkedItems.ToArray());
            SyncLinkAdditionalButton(false);
        }

        private void SyncLinkAdditionalButton(bool syncText = true)
        {
            if (linkedItems.Count == 0)
            {
                linkAdditionalButton.buttonColor = new Color(.839f, .839f, .839f, 1f);
                if(syncText) linkAdditionalButton.label = "Link Additional Items";
            }
            else
            {
                linkAdditionalButton.buttonColor = new Color(0.49f, 1f, 0.49f);
                if(syncText) linkAdditionalButton.label = $"Link Additional Items ({linkedItems.Count})";
            }
        }

        private void EquipItemByUid(string uid, bool val)
        {
            var geo = containingAtom.GetStorableByID("geometry");
            var package = uid.Substring(0, uid.IndexOf(":/"));
            var packageSplit = package.Split(new[] { "." }, StringSplitOptions.None);
            var packageBase = string.Join(".", packageSplit.Take(packageSplit.Length -1).ToArray());
            var version = int.Parse(packageSplit.Last());
            var startingVersion = 1;
            var nextUid = uid.Replace(package, $"{packageBase}.{startingVersion}");
            var toggle = geo.GetBoolJSONParam($"clothing:{nextUid}");
            if (toggle != null && toggle.val)
            {
                toggle.val = val;
            }
            else
            {
                JSONStorableBool latestToggle = null;
                for (int i = version; i < version+10; i++)
                {
                    nextUid = uid.Replace(package, $"{packageBase}.{i}");
                    toggle = geo.GetBoolJSONParam($"clothing:{nextUid}");
                    if(toggle != null)
                    {
                        latestToggle = toggle;
                    }
                }
                latestToggle.val = val;
            }
        }

        private void ToggleClothingItem(DAZClothingItem dci)
        {
            if (dci == null) return;
            var toggle = containingAtom.GetStorableByID("geometry").GetBoolJSONParam($"clothing:{dci.uid}");
            toggle.val = !toggle.val;
        }

        private void OnCollisionEnabledChanged(bool val)
        {
            lHeelCollider.enabled = val;
            rHeelCollider.enabled = val;
            lToeCollider.enabled = val;
            rToeCollider.enabled = val;
        }

        private void AdjustHeight()
        {
            if (heightAdjustMode.val == "None") return;
            StartCoroutine(OnHeightAdjustModeChanged(heightAdjustMode.val, false));
        }

        private IEnumerator OnHeightAdjustModeChanged(string val, bool disableCollision)
        {
            RestoreControllerHeights();
            if(disableCollision)
            {
                containingAtom.collisionEnabled = false;
                yield return new WaitForFixedUpdate();
            }
            switch (val)
            {
                case "None":
                {
                    currentHeightAdjustMode = 0;
                    break;
                }
                case "Feet":
                {
                    var delta = heelColliderScale.val.y - .06f * Mathf.Sin((footSlope.val) * Mathf.PI / 180f);
                    var pos = rfootCTRL.transform.position;
                    pos.y = rFootY + delta;
                    rfootCTRL.transform.position = pos;
                    
                    pos = lfootCTRL.transform.position;
                    pos.y = lFootY + delta;
                    lfootCTRL.transform.position = pos;
                    currentHeightAdjustMode = 1;
                    break;
                }
                case "Root":
                {
                    // StoreControllerHeights();
                    var delta = heelColliderScale.val.y - .06f * Mathf.Sin((footSlope.val) * Mathf.PI / 180f);
                    var pos = containingAtom.mainController.transform.position;
                    pos.y = rootY + delta;
                    containingAtom.mainController.transform.position = pos;
                    currentHeightAdjustMode = 2;
                    break;
                }
            }
            
            if(disableCollision)
            {
                yield return new WaitForSeconds(.5f);
                containingAtom.collisionEnabled = true;
            }
        }

        private void RegisterSliders(int i)
        {
            switch (i)
            {
                case 0:
                {
                    colScaleUid.RegisterVector(heelColliderScale);
                    colOffsetUid.RegisterVector(heelColliderOffset);
                    colRotationUid.RegisterVector(heelColliderRotation);
                    break;
                }
                case 1:
                {
                    colScaleUid.RegisterVector(toeColliderScale);
                    colOffsetUid.RegisterVector(toeColliderOffset);
                    colRotationUid.RegisterVector(toeColliderRotation);
                    break;
                }
            }
        }

        private void IndivToesSetLocked(bool val)
        {
            float[] defaultsY = { 15f, 10f, 10f, 10f, 20f };
            float[] defaultsZ = { 3f, 3f, 3f, 3f, 3f };
            foreach (var toe in toes)
            {
                var confJ = toe.GetComponent<ConfigurableJoint>();
                var hX = confJ.highAngularXLimit;
                var lX = confJ.lowAngularXLimit;
                var Y = confJ.angularYLimit;
                var Z = confJ.angularZLimit;
                if (val)
                {
                    hX.limit = .001f;
                    lX.limit = 0f;
                    Y.limit = Z.limit = .001f;
                }
                else
                {
                    hX.limit = 45f;
                    lX.limit = -50f;
                    Y.limit = defaultsY[toes.IndexOf(toe) % 5];
                    Z.limit = defaultsZ[toes.IndexOf(toe) % 5];
                }
                confJ.highAngularXLimit = hX;
                confJ.lowAngularXLimit = lX;
                confJ.angularYLimit = Y;
                confJ.angularZLimit = Z;
            }
        }

        private void ToeSetLocked(bool val)
        {
            // This was adapted from Stopper' Contortionist plugin.
            var confJR = rToe.GetComponent<ConfigurableJoint>();
            SoftJointLimit[] hLimR = { confJR.highAngularXLimit, confJR.angularYLimit, confJR.angularZLimit };
            var lLimR = confJR.lowAngularXLimit;
            
            var confJL = lToe.GetComponent<ConfigurableJoint>();
            SoftJointLimit[] hLimL = { confJL.highAngularXLimit, confJL.angularYLimit, confJL.angularZLimit };
            var lLimL = confJL.lowAngularXLimit;

            if (val)
            {
                var high = .5f * toeLimits.val + toeRotation.val;
                var low = -.5f * toeLimits.val + toeRotation.val;
                for (int i = 0; i < 3; i++)
                {
                    if(i==0) hLimR[i].limit = hLimL[i].limit = footSlope.val + high.x + .001f;
                    else hLimR[i].limit = hLimL[i].limit = high.x + .001f;
                }
                lLimR.limit = lLimL.limit = footSlope.val + low.x;
            }
            else
            {
                float[] hLims = { 75f, 8f, 15f };
                for (int i = 0; i < 3; i++)
                {
                    hLimR[i].limit = hLimL[i].limit = hLims[i];
                }
                lLimR.limit = lLimL.limit = -40f;
            }

            confJR.highAngularXLimit = hLimR[0];
            confJR.angularYLimit = hLimR[1];
            confJR.angularZLimit = hLimR[2];
            confJR.lowAngularXLimit = lLimR;
            
            confJL.highAngularXLimit = hLimL[0];
            confJL.angularYLimit = hLimL[1];
            confJL.angularZLimit = hLimL[2];
            confJL.lowAngularXLimit = lLimL;
            ReparentToeCollider(val);
            footSlope.SetInteractble(!val);
            footSlopeUid.label = val? "Foot Slope (unlock toe)" : footSlopeUid.label = "Foot Slope";
            if (mainTabbar.id == 1)
            {
                if(val)
                {
                    RemoveUIElements(UIElementsSliders);
                    lockedInfo = CreateTextField(new JSONStorableString("", "Unlock toe joint to edit the colliders."));
                    lockedInfo.height = 100f;
                    UIElements.Add(lockedInfo);
                }
                else CreateSliders();
            }
        }
        
        private void ReparentToeCollider(bool enabled)
        {
            switch (enabled)
            {
                case false:
                {
                    rToeCollider.transform.parent = rToeColROffset;
                    lToeCollider.transform.parent = lToeColROffset;
                    break;
                }
                case true:
                {
                    rToeCollider.transform.parent = rFoot.transform;
                    lToeCollider.transform.parent = lFoot.transform;
                    break;
                }
            }
            lToeCollisionHandler.enabled = rToeCollisionHandler.enabled = !enabled;
        }

        void StorePoseData()
        {
            rfootJointDrive = new Vector3(
                rfootCTRL.jointRotationDriveXTarget,
                rfootCTRL.jointRotationDriveYTarget,
                rfootCTRL.jointRotationDriveZTarget
            );
            lfootJointDrive = new Vector3(
                lfootCTRL.jointRotationDriveXTarget,
                lfootCTRL.jointRotationDriveYTarget,
                lfootCTRL.jointRotationDriveZTarget
            );
            rtoeJointDrive = new Vector3(
                rtoeCTRL.jointRotationDriveXTarget,
                rtoeCTRL.jointRotationDriveYTarget,
                rtoeCTRL.jointRotationDriveZTarget
            );
            ltoeJointDrive = new Vector3(
                ltoeCTRL.jointRotationDriveXTarget,
                ltoeCTRL.jointRotationDriveYTarget,
                ltoeCTRL.jointRotationDriveZTarget
            );
            rootY = containingAtom.mainController.transform.position.y;
            rFootY = rfootCTRL.transform.position.y;
            lFootY = lfootCTRL.transform.position.y;
        }

        private void RestoreJointDrives()
        {
            rfootCTRL.jointRotationDriveXTarget = rfootJointDrive.x;
            rfootCTRL.jointRotationDriveYTarget = rfootJointDrive.y;
            rfootCTRL.jointRotationDriveZTarget = rfootJointDrive.z;
            
            lfootCTRL.jointRotationDriveXTarget = lfootJointDrive.x;
            lfootCTRL.jointRotationDriveYTarget = lfootJointDrive.y;
            lfootCTRL.jointRotationDriveZTarget = lfootJointDrive.z;
            
            rtoeCTRL.jointRotationDriveXTarget = rtoeJointDrive.x;
            rtoeCTRL.jointRotationDriveYTarget = rtoeJointDrive.y;
            rtoeCTRL.jointRotationDriveZTarget = rtoeJointDrive.z;
            
            ltoeCTRL.jointRotationDriveXTarget = ltoeJointDrive.x;
            ltoeCTRL.jointRotationDriveYTarget = ltoeJointDrive.y;
            ltoeCTRL.jointRotationDriveZTarget = ltoeJointDrive.z;
        }
        
        private void RestoreControllerHeights()
        {
            switch (currentHeightAdjustMode)
            {
                case 1:
                {
                    var pos = rfootCTRL.transform.position;
                    pos.y = rFootY;
                    rfootCTRL.transform.position = pos;
            
                    pos = lfootCTRL.transform.position;
                    pos.y = lFootY;
                    lfootCTRL.transform.position = pos;
                    break;
                }
                case 2:
                {
                    var pos = root.transform.position;
                    pos.y = rootY;
                    root.transform.position = pos;
                    break;
                }
            }
        }
        
        private void Reset()
        {
            if (!autoCalculateDeltas.val) lDelta = rDelta = 0f;
            else rFootSlopeDelta.val = lFootSlopeDelta.val = 0f;
            toeRotation.SetValToDefault();
            footSlope.SetValToDefault();
            lockToes.val = false;
            lockIndividualToes.val = false;
            RestoreControllerHeights();
            heelColliderOffset.SetValToDefault();
            heelColliderRotation.SetValToDefault();
            heelColliderScale.SetValToDefault();
            toeColliderOffset.SetValToDefault();
            toeColliderRotation.SetValToDefault();
            toeColliderScale.SetValToDefault();
            rToeGPUCapsule.enabled = lToeGPUCapsule.enabled = lHeelGPUCapsule.enabled = rHeelGPUCapsule.enabled = false;
            reloadButton.label = reloadLabel = "<b>Toggle</b>: None";
            if (!autoCalculateDeltas.val)
            {
                rDelta = rFootSlopeDelta.val;
                lDelta = lFootSlopeDelta.val;
            }
        }

        private void ZeroJointDrives()
        {
            return;
            rfootCTRL.jointRotationDriveXTarget =
                rfootCTRL.jointRotationDriveYTarget = rfootCTRL.jointRotationDriveZTarget = 0f;
            lfootCTRL.jointRotationDriveXTarget =
                lfootCTRL.jointRotationDriveYTarget = lfootCTRL.jointRotationDriveZTarget = 0f;
            rtoeCTRL.jointRotationDriveXTarget =
                rtoeCTRL.jointRotationDriveYTarget = rtoeCTRL.jointRotationDriveZTarget = 0f;
            ltoeCTRL.jointRotationDriveXTarget =
                ltoeCTRL.jointRotationDriveYTarget = ltoeCTRL.jointRotationDriveZTarget = 0f;
        }

        private TriggerHandler rToeTriggerHandler;
        private TriggerHandler lToeTriggerHandler;

        private void AddToeColliders()
        {
            rToeCollisionHandler = rToe.GetComponent<Rigidbody>().gameObject.AddComponent<CollisionHandler>();
            lToeCollisionHandler = lToe.GetComponent<Rigidbody>().gameObject.AddComponent<CollisionHandler>();
            rToeColliderBase = new GameObject().transform;
            rToeColliderBase.parent = rToe.transform;
            rToeColliderBase.localPosition = Vector3.zero;
            rToeColliderBase.localRotation = Quaternion.Euler(rToeBaseRotation);
            rToeColliderBase.Translate(new Vector3(0f, -.02f, 0f));
            rToeColliderBase.name = "rToeColliderBase";

            rToeColTOffset.parent = rToeColliderBase;
            rToeColTOffset.localPosition = Vector3.zero;
            rToeColTOffset.localRotation = Quaternion.identity;
            
            rToeColROffset.parent = rToeColTOffset;
            rToeColROffset.localPosition = Vector3.zero;
            rToeColROffset.localRotation = Quaternion.identity;

            rToeCollider = GameObject.CreatePrimitive(PrimitiveType.Cube).GetComponent<BoxCollider>();
            rToeCollider.name = "rToeCollider";
            rToeTriggerHandler = AddBoxTrigger(rToeCollider).gameObject.AddComponent<TriggerHandler>();
            rToeCollisionHandler.audioSource = rToeCollider.gameObject.AddComponent<AudioSource>();
            rToeCollisionHandler.audioSource.spatialize = true;
            rToeCollisionHandler.audioSource.spatialBlend = 1f;
            var transform1 = rToeCollider.transform;
            transform1.parent = rToeColROffset;
            transform1.localScale = toeColliderScale.val;
            transform1.localPosition = new Vector3(0f, -toeColliderScale.val.y * .5f, 0f);
            transform1.localRotation = new Quaternion();
            rToeCollider.enabled = true;
            var material = rToeCollider.GetComponent<Renderer>().material;
            material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
            material.color = ShoeColliderColor;
            foreach (var col in rToe.transform.GetComponentsInChildren<Collider>())
            {
                Physics.IgnoreCollision(rToeCollider, col);
            }
            rToeGPUCapsule = rToeColROffset.gameObject.AddComponent<CapsuleLineSphereCollider>();
            GPUCollidersManager.RegisterLineSphereCollider(rToeGPUCapsule);
            rToeGPUCapsule.enabled = false;
            
            
            lToeColliderBase = new GameObject().transform;
            lToeColliderBase.parent = lToe.transform;
            lToeColliderBase.localPosition = Vector3.zero;
            lToeColliderBase.localRotation = Quaternion.Euler(lToeBaseRotation);
            lToeColliderBase.Translate(new Vector3(0f, -.02f, 0f));
            lToeColliderBase.name = "lToeColliderBase";

            lToeColTOffset.parent = lToeColliderBase;
            lToeColTOffset.localPosition = Vector3.zero;
            lToeColTOffset.localRotation = Quaternion.identity;
            
            lToeColROffset.parent = lToeColTOffset;
            lToeColROffset.localPosition = Vector3.zero;
            lToeColROffset.localRotation = Quaternion.identity;
            
            lToeCollider = GameObject.CreatePrimitive(PrimitiveType.Cube).GetComponent<BoxCollider>();
            lToeCollider.name = "lToeCollider";
            lToeTriggerHandler = AddBoxTrigger(rToeCollider).gameObject.AddComponent<TriggerHandler>();
            lToeCollisionHandler.audioSource = lToeCollider.gameObject.AddComponent<AudioSource>();
            lToeCollisionHandler.audioSource.spatialize = true;
            lToeCollisionHandler.audioSource.spatialBlend = 1f;
            transform1 = lToeCollider.transform;
            transform1.parent = lToeColROffset;
            transform1.localScale = toeColliderScale.val;
            transform1.localPosition = new Vector3(0f, -toeColliderScale.val.y * .5f, 0f);
            transform1.localRotation = new Quaternion();
            lToeCollider.enabled = true;
            material = lToeCollider.GetComponent<Renderer>().material;
            material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
            material.color = ShoeColliderColor;
            foreach (var col in lToe.transform.GetComponentsInChildren<Collider>())
            {
                Physics.IgnoreCollision(lToeCollider, col);
            }
            lToeGPUCapsule = lToeColROffset.gameObject.AddComponent<CapsuleLineSphereCollider>();
            GPUCollidersManager.RegisterLineSphereCollider(lToeGPUCapsule);
            lToeGPUCapsule.enabled = false;
        }
        
        private void AddHeelColliders()
        {
            lHeelCollisionHandler = rFoot.GetComponent<Rigidbody>().gameObject.AddComponent<CollisionHandler>();
            lHeelCollisionHandler.triggerHandler = lToeTriggerHandler;
            rHeelCollisionHandler = lFoot.GetComponent<Rigidbody>().gameObject.AddComponent<CollisionHandler>();
            rHeelCollisionHandler.triggerHandler = rToeTriggerHandler;
            rHeelColliderBase = new GameObject().transform;
            rHeelColliderBase.parent = rFoot.transform;
            rHeelColliderBase.localPosition = Vector3.zero;
            rHeelColliderBase.localRotation = Quaternion.Euler(rHeelBaseRotation);
            rHeelColliderBase.Translate(new Vector3(0f, -.06f, -.025f));
            rHeelColliderBase.name = "rHeelColliderBase";

            rHeelColTOffset.parent = rHeelColliderBase;
            rHeelColTOffset.localPosition = Vector3.zero;
            rHeelColTOffset.localRotation = Quaternion.identity;

            rHeelColROffset.parent = rHeelColTOffset;
            rHeelColROffset.localPosition = Vector3.zero;
            rHeelColROffset.localRotation = Quaternion.identity;

            rHeelCollider = GameObject.CreatePrimitive(PrimitiveType.Cube).GetComponent<BoxCollider>();
            rHeelCollider.name = "rHeelCollider";
            AddBoxTrigger(rHeelCollider);
            rHeelCollisionHandler.audioSource = rHeelCollider.gameObject.AddComponent<AudioSource>();
            rHeelCollisionHandler.audioSource.spatialize = true;
            rHeelCollisionHandler.audioSource.spatialBlend = 1f;
            var transform1 = rHeelCollider.transform;
            transform1.parent = rHeelColROffset;
            transform1.localScale = heelColliderScale.val;
            transform1.localPosition = new Vector3(0f, -heelColliderScale.val.y * .5f, 0f);
            transform1.localRotation = new Quaternion();
            
            rHeelCollider.enabled = true;
            var material = rHeelCollider.GetComponent<Renderer>().material;
            material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
            material.color = ShoeColliderColor;
            foreach (var col in rToe.transform.GetComponentsInChildren<Collider>())
            {
                Physics.IgnoreCollision(rHeelCollider, col);
            }
            rHeelGPUCapsule = rHeelColROffset.gameObject.AddComponent<CapsuleLineSphereCollider>();
            GPUCollidersManager.RegisterLineSphereCollider(rHeelGPUCapsule);
            rHeelGPUCapsule.enabled = false;
            

            lHeelColliderBase = new GameObject().transform;
            lHeelColliderBase.parent = lFoot.transform;
            lHeelColliderBase.localPosition = Vector3.zero;
            lHeelColliderBase.localRotation = Quaternion.Euler(lHeelBaseRotation);
            lHeelColliderBase.Translate(new Vector3(0f, -.06f, -.025f));
            lHeelColliderBase.name = "lHeelColliderBase";

            lHeelColTOffset.parent = lHeelColliderBase;
            lHeelColTOffset.localPosition = Vector3.zero;
            lHeelColTOffset.localRotation = Quaternion.identity;

            lHeelColROffset.parent = lHeelColTOffset;
            lHeelColROffset.localPosition = Vector3.zero;
            lHeelColROffset.localRotation = Quaternion.identity;

            lHeelCollider = GameObject.CreatePrimitive(PrimitiveType.Cube).GetComponent<BoxCollider>();
            lHeelCollider.name = "lHeelCollider";
            AddBoxTrigger(lHeelCollider);
            lHeelCollisionHandler.audioSource = lHeelCollider.gameObject.AddComponent<AudioSource>();
            lHeelCollisionHandler.audioSource.spatialize = true;
            lHeelCollisionHandler.audioSource.spatialBlend = 1f;
            transform1 = lHeelCollider.transform;
            transform1.parent = lHeelColROffset;
            transform1.localScale = heelColliderScale.val;
            transform1.localPosition = new Vector3(0f, -heelColliderScale.val.y * .5f, 0f);
            transform1.localRotation = new Quaternion();

            lHeelCollider.enabled = true;
            material = lHeelCollider.GetComponent<Renderer>().material;
            material.shader = Shader.Find("Battlehub/RTGizmos/Handles");
            material.color = ShoeColliderColor;
            foreach (var col in lToe.transform.GetComponentsInChildren<Collider>())
            {
                Physics.IgnoreCollision(lHeelCollider, col);
            }

            lHeelGPUCapsule = lHeelColROffset.gameObject.AddComponent<CapsuleLineSphereCollider>();
            GPUCollidersManager.RegisterLineSphereCollider(lHeelGPUCapsule);
            lHeelGPUCapsule.enabled = false;
        }

        private BoxCollider AddBoxTrigger(BoxCollider collider)
        {
            BoxCollider trigger = GameObject.CreatePrimitive(PrimitiveType.Cube).GetComponent<BoxCollider>();
            // trigger.enabled = false;
            trigger.isTrigger = true;
            var transform1 = trigger.transform;
            transform1.parent = collider.transform;
            transform1.localRotation = Quaternion.identity;
            transform1.localPosition = Vector3.zero;
            transform1.localScale = new Vector3(1.4f ,1.3f, 1.4f);
            trigger.gameObject.name = collider.name + "Trigger";
            trigger.GetComponent<Renderer>().enabled = false;
            var foot = collider.name.StartsWith("l") ? lFoot : rFoot;
            foreach (var col in foot.transform.GetComponentsInChildren<Collider>())
            {
                Physics.IgnoreCollision(trigger, col);
            }
            Physics.IgnoreCollision(trigger, collider);
            return trigger;
        }

        private CapsuleLineSphereCollider lHeelGPUCapsule;
        private CapsuleLineSphereCollider lToeGPUCapsule;
        private CapsuleLineSphereCollider rHeelGPUCapsule;
        private CapsuleLineSphereCollider rToeGPUCapsule;


        private float zeta;
        private float sin;
        
        private void OnHeelScaleChanged(Vector3 val, bool rebase = true)
        {
            if (!rHeelColliderBase) return;
            var transform1 = rHeelCollider.transform;
            var transform2 = lHeelCollider.transform;
            transform1.localScale = transform2.localScale = val;
            transform1.localPosition = transform2.localPosition = new Vector3(0f, -.5f * val.y, val.z * .5f);
            var r = Mathf.Sqrt(val.x *val.x + val.z * val.z);
            lHeelGPUCapsule.RadiusA = lHeelGPUCapsule.RadiusB =
                rHeelGPUCapsule.RadiusA = rHeelGPUCapsule.RadiusB = r;
            lHeelGPUCapsule.B = rHeelGPUCapsule.B = new Vector3(0f, .5f * r - val.y, 0f);
            if(rebase && sin > 0f) zeta = heelColliderScale.val.y / sin + heelColliderOffset.val.z;
            AdjustHeight();
            // lHeelCollisionHandler.area = rHeelCollisionHandler.area = val.x * val.z;
        }
        
        private void OnHeelOffsetChanged(Vector3 val)
        {
            if (!rHeelColliderBase) return;
            rHeelColTOffset.localPosition = val;
            val.x *= -1f;
            lHeelColTOffset.transform.localPosition = val;
            var scale = heelColliderScale.val;
            if(sin > 0) scale.y = (zeta - val.z) * sin;
            heelColliderScale.valNoCallback = scale;
            OnHeelScaleChanged(scale, false);
        }
        
        private void OnHeelRotationChanged(Vector3 val)
        {
            if (!rHeelColliderBase) return;
            rHeelColROffset.localRotation = Quaternion.Euler(new Vector3(-footSlope.val, 0f, 0f)) * Quaternion.Euler(val);
            val.y *= -1f;
            val.z *= -1f;
            lHeelColROffset.localRotation = Quaternion.Euler(new Vector3(-footSlope.val, 0f, 0f)) * Quaternion.Euler(val);
        }

        private void OnToeScaleChanged(Vector3 val)
        {
            if (!rToeColliderBase) return;
            var transform1 = rToeCollider.transform;
            var transform2 = lToeCollider.transform;
            transform1.localScale = transform2.localScale = val;
            transform1.localPosition = transform2.localPosition = new Vector3(0f, -.5f * val.y, 0f);
            var r = .5f * Mathf.Sqrt(val.x * val.x + val.z * val.z);
            lToeGPUCapsule.B = rToeGPUCapsule.B = new Vector3(0f, .75f * r - val.y, 0f);
            lToeGPUCapsule.RadiusA = lToeGPUCapsule.RadiusB = rToeGPUCapsule.RadiusA = rToeGPUCapsule.RadiusB = r;
            // lToeCollisionHandler.area = rToeCollisionHandler.area = val.x * val.z;
        }
        
        private void OnToeOffsetChanged(Vector3 val)
        {
            if (!rToeColliderBase) return;
            rToeColTOffset.localPosition = val;
            val.x *= -1f;
            lToeColTOffset.localPosition = val;
        }
        
        private void OnToeRotationChanged(Vector3 val)
        {
            if (!rToeColliderBase) return;
            rToeColROffset.localRotation = Quaternion.Euler(val);
            val.y *= -1f;
            val.z *= -1f;
            lToeColROffset.localRotation = Quaternion.Euler(val);
        }

        private void OnFootPoseChanged()
        {
            SetFootPose(footSlope.val);
        }

        private void SetFootPose(float val)
        {
            Vector3 rt = new Vector3(toeRotation.val.x + footSlope.val, toeRotation.val.y, toeRotation.val.z);
            Vector3 lt = new Vector3(toeRotation.val.x + footSlope.val, -toeRotation.val.y, -toeRotation.val.z);
            OnHeelRotationChanged(heelColliderRotation.val);
            // OnToeRotationChanged(toeColliderRotation.val);
            if(!rToeGPUCapsule.enabled) rToeGPUCapsule.enabled = lToeGPUCapsule.enabled = lHeelGPUCapsule.enabled = rHeelGPUCapsule.enabled = true;
            var lFootX = -val - lDelta;
            var rFootX = -val - rDelta;
            lFootShinLink.targetRotation = lFootShinLink.targetRotation.SetX(lFootX);
            lFootCtrlLink.targetRotation = lFootCtrlLink.targetRotation.SetX(lFootX);
            rFootShinLink.targetRotation = rFootShinLink.targetRotation.SetX(rFootX);
            rFootCtrlLink.targetRotation = rFootCtrlLink.targetRotation.SetX(rFootX);
            
            lToeFootLink.targetRotation = Quaternion.Euler(lt);
            rToeFootLink.targetRotation = Quaternion.Euler(rt);
            
            sin = Mathf.Sin((footSlope.val) * Mathf.PI / 180f);
            if(sin > 0f) zeta = heelColliderScale.val.y / sin + heelColliderOffset.val.z;
        }
        
        private IEnumerator IECalculateDeltas()
        {
            lDelta = rDelta = 0f;
            var col = collisionEnabled.val;
            collisionEnabled.val = false;
            SetFootPose(0f);
            yield return new WaitForSeconds(.5f);
            CalculateDeltas();
            OnFootPoseChanged();
            collisionEnabled.val = col;
        }

        private void CalculateDeltasFromJSON()
        {
            if (pm.lastLoadedJSON == null || !autoCalculateDeltas.val) return;
            float rX = 0f;
            float lX = 0f;
            foreach (JSONClass jc in pm.lastLoadedJSON["storables"].AsArray)
            {
                if ((string)jc["id"] == "lFoot")
                {
                    lX = jc["rotation"].AsObject["x"].AsFloat;
                }
                else if ((string)jc["id"] == "rFoot")
                {
                    rX = jc["rotation"].AsObject["x"].AsFloat;
                }
            }
            rFootSlopeDelta.val = rDelta = GetDelta(rX);
            lFootSlopeDelta.val = lDelta = GetDelta(lX);
        }

        private void CalculateDeltas()
        {
            rFootSlopeDelta.val = rDelta = GetDelta(rFoot.transform.localRotation.eulerAngles.x);
            lFootSlopeDelta.val = lDelta = GetDelta(lFoot.transform.localRotation.eulerAngles.x);
        }

        private float GetDelta(float rotation)
        {
            if (footSlope.val == 0f || rotation > 180f) return 0f;
            if (rotation + footSlope.val > footSlopeMax.val)
            {
                return Mathf.Max(footSlopeMax.val - rotation - footSlope.val, - footSlope.val);
            }
            return 0f;
        }
        
        private void OnClothingItemToggled(DAZClothingItem dci, bool equipped)
        {
            string itemName = StripNonAlphaNum(dci.creatorName) + "-" + StripNonAlphaNum(GetItemNameFromCustomClothingUID(dci.uid));
            var activeItemNames = new List<string>(itemToStore.choices);

            string preset = existingPresets.FirstOrDefault(x => x.Contains(itemName));
            if(equipped)
            {
                activeItems[itemName] = dci;
                if (!activeItemNames.Contains(itemName)) activeItemNames.Insert(1, itemName);
                if (preset != null)
                {
                    toggleItemButton.label = toggleItemLabel = $"<b>Toggle</b> {StripNonAlphaNum(GetItemNameFromCustomClothingUID(dci.uid))}";
                    if(enabled)
                    {
                        LoadFromPath(preset);
                        if(autoCalculateDeltas.val)
                        {
                            if (currentPreset != null) StartCoroutine(IECalculateDeltas());
                            else CalculateDeltas();
                        }
                    }
                    currentPreset = preset;
                    currentItem = dci;
                }
            }
            else
            {
                activeItems.Remove(itemName);
                activeItemNames.Remove(itemName);
                if (currentPreset == null || currentPreset == preset)
                {
                    if(currentPreset != null) EquipLinkedItems(LoadJSON(currentPreset).AsObject, false);
                    currentPreset = null;
                    // currentItem = null;
                    // toggleItemButtonUid.label = "Current item: None"
                    Reset();
                }
            }
            itemToStore.choices = activeItemNames;
            aditionalItemChooser.choices = activeItemNames;
        }

        private void HandShakeWithFlexIt()
        {
            Transform flexIt = null;
            for (int i = 0; i < transform.parent.childCount; i++)
            {
                var child = transform.parent.GetChild(i);
                if (child.name.EndsWith("CheesyFX.FlexIt"))
                {
                    child.SendMessage("HandShakeWithHeelAdjust", SendMessageOptions.DontRequireReceiver);
                    break;
                }
            }
        }

        private string GetItemNameFromCustomClothingUID(string uid)
        {
            string itemName = uid;
            if (uid.Contains("/"))
            {
                itemName = uid.Substring(uid.LastIndexOf("/") + 1);
                if (uid.EndsWith(".vam"))
                {
                    itemName = itemName.Substring(0, itemName.Length - 4);
                }
            }
            return (itemName);
        }

        private static string StripNonAlphaNum(string s)
        {
            return Regex.Replace(s, @"[^A-Za-z0-9]+", "");
        }
        
        private void SavePreset(string itemName)
        {
            if (itemName == "None")
            {
                SuperController.LogError(
                    "ERROR HeelAdjust2: No preset saved!\n First, select an active clothing item to store the preset for from the dropdown.");
                return;
            }

            var path = saveDir + "/" + itemName + ".json";
            SaveJSON(Store(), path);
            existingPresets.Insert(0, path);
            ($"Preset saved for '{itemName}'").Print();
        }

        private JSONClass Store()
        {
            var jc = new JSONClass();
            footSlope.Store(jc);
            toeRotation.Store(jc);
            heelColliderScale.Store(jc);
            heelColliderOffset.Store(jc);
            heelColliderRotation.Store(jc);
            toeColliderScale.Store(jc);
            toeColliderOffset.Store(jc);
            toeColliderRotation.Store(jc);
            toeLimits.Store(jc);
            lockToes.Store(jc);
            lockIndividualToes.Store(jc);
            JSONArray ja = new JSONArray();
            if(linkedItems.Count > 0)
            {
                for (int i = 0; i < linkedItems.Count; i++)
                {
                    ja[i] = activeItems[linkedItems[i]].uid;
                }

                jc["LinkedItems"] = ja;
            }
            return jc;
        }
        
        private void LoadFromJSON(JSONClass jc)
        {
            lockToes.val = false;
            footSlope.Load(jc);
            toeRotation.Load(jc);
            heelColliderOffset.Load(jc);
            heelColliderRotation.Load(jc);
            heelColliderScale.Load(jc);
            toeColliderScale.Load(jc);
            toeColliderOffset.Load(jc);
            toeColliderRotation.Load(jc);
            toeLimits.Load(jc);
            lockIndividualToes.Load(jc);
            EquipLinkedItems(jc, true);

            lFootSlopeDelta.max = 90f - footSlope.val;
            lFootSlopeDelta.min = -90f - footSlope.val;
            rFootSlopeDelta.max = 90f - footSlope.val;
            rFootSlopeDelta.min = -90f - footSlope.val;

            footSlopeMax.min = footSlope.val;
            StartCoroutine(IELockToe(jc));
        }

        private void EquipLinkedItems(JSONClass jc, bool val)
        {
            if (!jc.HasKey("LinkedItems")) return;
            var ja = jc["LinkedItems"].AsArray;
            foreach (JSONNode uid in ja)
            {
                EquipItemByUid(uid, val);
            }
        }

        private IEnumerator IELockToe(JSONClass jc)
        {
            lToeCollider.enabled = rToeCollider.enabled = false;
            lHeelCollider.enabled = rHeelCollider.enabled = false;
            yield return new WaitForSeconds(.5f);
            lockToes.Load(jc);
            lToeCollider.enabled = rToeCollider.enabled = true;
            lHeelCollider.enabled = rHeelCollider.enabled = true;
        }

        private void LoadFromPath(string path)
        {
            if (path.EndsWith(".json")) LoadFromJSON(LoadJSON(path).AsObject);
            else if (path.EndsWith(".heeladjust")) LegacyLoad(path);
            AdjustHeight();
            reloadButton.label= reloadLabel = "<b>Reload</b> "+path.Split(new string[] { "\\" }, StringSplitOptions.None).Last();
        }

        private void LegacyLoad(string path)
        {
            string loadedSettings = SuperController.singleton.ReadFileIntoString(path);
            string[] loadedSettingsLines = loadedSettings.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
            string[] footVals = loadedSettingsLines[0].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeVals = loadedSettingsLines[1].Split(new string[] { "," }, StringSplitOptions.None);
            string[] heelColliderVals = loadedSettingsLines[2].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeColliderVals = loadedSettingsLines[3].Split(new string[] { "," }, StringSplitOptions.None);

            toeRotation.valNoCallback = new Vector3(float.Parse(toeVals[0]), float.Parse(toeVals[1]), float.Parse(toeVals[2]));
            footSlope.valNoCallback = float.Parse(footVals[0]);
            OnFootPoseChanged();
            var h = float.Parse(heelColliderVals[0])*.01f;
            var z = float.Parse(heelColliderVals[1]);

            Transform sole = new GameObject().transform;
            sole.rotation = Quaternion.Euler(footSlope.val, 0f, 0f);
            sole.Translate(new Vector3(0f, -.06f, -.025f));
            var v = sole.InverseTransformPoint(new Vector3(0f, -h, z));
            heelColliderOffset.val = new Vector3(0f, 0f, -v.z - v.y * Mathf.Tan(footSlope.val * Mathf.PI/180f));
            heelColliderScale.val = new Vector3(.02f,  -v.y / Mathf.Cos(footSlope.val * Mathf.PI/180f), .02f);
            toeColliderScale.val = new Vector3(.05f, float.Parse(toeColliderVals[0])*.01f - .02f, .05f);
            
            Destroy(sole.gameObject);
        }

        private void GetExistingPresets()
        {
            existingPresets = SuperController.singleton.GetFilesAtPath(saveDir).ToList();
            var packagePath = this.GetPackagePath();
            if (packagePath != String.Empty)
            {
                foreach (var file in SuperController.singleton.GetFilesAtPath(packagePath+saveDir))
                {
                    if (!existingPresets.Contains(file.Replace(packagePath, ""))) existingPresets.Add(file);
                }
            }
            if(FileManagerSecure.DirectoryExists("Saves/HeelAdjust"))
                existingPresets.AddRange(SuperController.singleton.GetFilesAtPath("Saves/HeelAdjust"));
        }

        private Request footstepBundle;
        public void LoadBundle(){
            Request request = new AssetLoader.AssetBundleFromFileRequest{
                path = bundlePath,
                callback = OnBundleLoaded};
            AssetLoader.QueueLoadAssetBundleFromFile(request);
        }
		
        private void OnBundleLoaded(Request request){
            try{
                footstepBundle = request;
                var libraries = new List<string>();
                request.assetBundle.GetAllAssetNames().
                    Where(x => x.EndsWith(".ogg")||x.EndsWith(".wav")||x.EndsWith(".mp3")).ToList().
                    ForEach(x =>
                    {
                        var library = x.Split('/')[2];
                        if(!libraries.Contains(library)) libraries.Add(library);
                    });
                soundLibrary.choices = libraries;
                GetClipLibraryFromAssetBundle(request, "heels");
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void GetClipLibraryFromAssetBundle(Request request, string library)
        {
            var paths = request.assetBundle.GetAllAssetNames().
                Where(x => x.EndsWith(".ogg")||x.EndsWith(".wav")||x.EndsWith(".mp3")).
                Where(x => x.Split('/')[2] == library);
            AudioManager.clips = paths.Select(x => request.assetBundle.LoadAsset<AudioClip>(x)).ToList();
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
            jc["HeelAdjust2"] = Store();
            return jc;
        }
        
        public override void LateRestoreFromJSON(
            JSONClass jc,
            bool restorePhysical,
            bool restoreAppearance,
            bool setMissingToDefault
        )
        {
            if (jc.HasKey("HeelAdjust2"))
            {
                sceneData = jc["HeelAdjust2"].AsObject;
            }
        }
        
        public float rightUIHeight
        {
            get
            {
                float height = 0f;
                for (int i = 0; i < rightUIElements.Count; i++)
                {
                    var le = rightUIElements[i].transform.GetComponent<LayoutElement>();
                    height += le.preferredHeight;
                }
                return height;
            }
        }

        public float leftUIHeight
        {
            get
            {
                float height = 5f;
                for (int i = 0; i < leftUIElements.Count; i++)
                {
                    var le = leftUIElements[i].transform.GetComponent<LayoutElement>();
                    height += le.preferredHeight;
                }
                return height;
            }
        }

        private static GameObject vector3SliderPrefab;
        public UIDynamicV3Slider CreateV3Slider(MyJSONStorableVector3 vector = null)
        {
            if (vector3SliderPrefab == null) vector3SliderPrefab = CreateVector3SliderPrefab(this);
           
            Transform t = CreateUIElement(vector3SliderPrefab.transform, false);
            var uid = t.GetComponent<UIDynamicV3Slider>();
            if (vector != null) uid.RegisterVector(vector);
            t.gameObject.SetActive(true);
            float spacerHeight = leftUIHeight - rightUIHeight - 5f;
            var spacer = CreateSpacer(true);
            spacer.height = spacerHeight;
            uid.spacer = spacer;
            return uid;
        }
        
        private static GameObject CreateVector3SliderPrefab(MVRScript script)
        {
            GameObject prefab = new GameObject("Vector3Slider");
            prefab.SetActive(false);
            UIDynamicV3Slider uiDynamicV3Slider = prefab.AddComponent<UIDynamicV3Slider>();
            RectTransform rt = prefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(1064, 0);
            rt.offsetMin = new Vector2(10, 0);

            LayoutElement le = prefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = 130;
            le.minWidth = 1064;
            le.preferredHeight = 130;
            le.preferredWidth = 1064;

            RectTransform backgroundTransform =
                script.manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform =
                Instantiate(backgroundTransform, prefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            backgroundTransform.offsetMax = new Vector2(0, -2.5f);
            backgroundTransform.offsetMin = new Vector2(0, 2.5f);

            float w = .33f;
            float d = 0f;
            for (int k = 0; k < 3; k++)
            {
                rt = Instantiate(
                    script.manager.configurableSliderPrefab.transform,
                    prefab.transform,
                    false) as RectTransform;
                var uiDynamicSlider = rt.GetComponent<UIDynamicSlider>();
                uiDynamicSlider.sliderControl.clamp = false;
                uiDynamicV3Slider.sliders.Add(uiDynamicSlider);
                rt.anchorMax = new Vector2(k * w + .33f, 1);
                rt.anchorMin = new Vector2(k * (w) + d, 1);
                rt.offsetMax = new Vector2(-4, -10);
                rt.offsetMin = new Vector2(4f, -120);
                rt.gameObject.SetActive(true);
            }
            return prefab;
        }
        
        public UIDynamicTabBar CreateTabBar(string[] menuItems, Action<int> selectTab, int columns = -1,
           bool invokeDefault = false, float width = 1f, float itemHeight = 50f)
        {
            if (columns == -1) columns = menuItems.Length;
            float rowCount = Mathf.Ceil((float)menuItems.Length / columns);
            float totalHeight = 10f + itemHeight * (rowCount);
            UIDynamicTabBar tabBar;
            GameObject tabbarPrefab = new GameObject("TabBar");
            RectTransform rt = tabbarPrefab.AddComponent<RectTransform>();
            rt.anchorMax = new Vector2(0, 1);
            rt.anchorMin = new Vector2(0, 1);
            rt.offsetMax = new Vector2(1064, -500);
            rt.offsetMin = new Vector2(10, -600);
            LayoutElement le = tabbarPrefab.AddComponent<LayoutElement>();
            le.flexibleWidth = 1;
            le.minHeight = totalHeight;
            le.minWidth = 1064;
            le.preferredHeight = totalHeight;
            le.preferredWidth = 1064;

            RectTransform backgroundTransform =
                manager.configurableScrollablePopupPrefab.transform.Find("Background") as RectTransform;
            backgroundTransform = Instantiate(backgroundTransform, tabbarPrefab.transform);
            backgroundTransform.name = "Background";
            backgroundTransform.anchorMax = new Vector2(1, 1);
            backgroundTransform.anchorMin = new Vector2(0, 0);
            //backgroundTransform.offsetMax = new Vector2(0, -15);
            //backgroundTransform.offsetMin = new Vector2(0, 15);

            UIDynamicTabBar uid = tabbarPrefab.AddComponent<UIDynamicTabBar>();
            float padding = 5.0f;
            float itemWidth, rowOffset;
            itemWidth = 1044f * width / columns - padding;

            float x = 15;
            for (int i = 0; i < menuItems.Length; ++i)
            {
                int rowNumber = i / columns;
                rowOffset = itemHeight * rowNumber;
                if (i % columns == 0) x = 15f;
                RectTransform buttonTransform =
                    manager.configurableScrollablePopupPrefab.transform.Find("Button") as RectTransform;
                buttonTransform = Instantiate(buttonTransform, tabbarPrefab.transform);
                buttonTransform.name = "Button";
                buttonTransform.anchorMax = new Vector2(0, 1);
                buttonTransform.anchorMin = new Vector2(0, 1);
                buttonTransform.offsetMax = new Vector2(x + itemWidth, -10 - rowOffset);
                buttonTransform.offsetMin = new Vector2(x, -itemHeight - rowOffset);
                Button buttonButton = buttonTransform.GetComponent<Button>();
                uid.buttons.Add(buttonButton);
                Text buttonText = buttonTransform.Find("Text").GetComponent<Text>();
                buttonText.text = menuItems[i];
                x += itemWidth + padding;
            }

            Transform t = CreateUIElement(tabbarPrefab.transform, false);
            tabBar = t.gameObject.GetComponent<UIDynamicTabBar>();
            for (int i = 0; i < tabBar.buttons.Count; ++i)
            {
                int menuID = i;
                tabBar.buttons[i].onClick.AddListener(() =>
                {
                    tabBar.lastId = tabBar.id;
                    tabBar.id = menuID;
                    for (int j = 0; j < tabBar.buttons.Count; ++j) tabBar.buttons[j].interactable = (j != menuID);
                    selectTab(menuID);
                });
                ColorBlock cb = ColorBlock.defaultColorBlock;
                cb.disabledColor = new Color(0.55f, 0.90f, 1f);
                cb.highlightedColor *= .8f;
                tabBar.buttons[i].colors = cb;
            }

            Destroy(tabbarPrefab);
            UIDynamic spacer = CreateSpacer(true);
            spacer.height = totalHeight;
            tabBar.spacer = spacer;
            tabBar.SelectTab = val => tabBar.buttons[val].onClick.Invoke();
            tabBar.SelectLast = () => tabBar.SelectTab(tabBar.id);
            if (invokeDefault) tabBar.SelectTab(tabBar.id);
            return tabBar;
        }
        
        public void RemoveUIElements(List<object> UIElements)
		{
			for (int i=0; i<UIElements.Count; ++i)
			{
				if (UIElements[i] is JSONStorableParam)
				{
					JSONStorableParam jsp = UIElements[i] as JSONStorableParam;
					if (jsp is JSONStorableFloat)
						RemoveSlider(jsp as JSONStorableFloat);
					else if (jsp is JSONStorableBool)
						RemoveToggle(jsp as JSONStorableBool);
					else if (jsp is JSONStorableColor)
						RemoveColorPicker(jsp as JSONStorableColor);
					else if (jsp is JSONStorableString)
						RemoveTextField(jsp as JSONStorableString);
					else if (jsp is JSONStorableStringChooser)
					{
						// Workaround for VaM not cleaning its panels properly.
						JSONStorableStringChooser jssc = jsp as JSONStorableStringChooser;
						RectTransform popupPanel = jssc.popup?.popupPanel;
						RemovePopup(jssc);
						if (popupPanel != null)
							UnityEngine.Object.Destroy(popupPanel.gameObject);
					}
				}
				else if (UIElements[i] is UIDynamic)
				{
					UIDynamic uid = UIElements[i] as UIDynamic;
					if (uid is UIDynamicButton)
						RemoveButton(uid as UIDynamicButton);
                    else if (uid is UIDynamicSlider)
						RemoveSlider(uid as UIDynamicSlider);
					else if (uid is UIDynamicToggle)
						RemoveToggle(uid as UIDynamicToggle);
					else if (uid is UIDynamicColorPicker)
						RemoveColorPicker(uid as UIDynamicColorPicker);
					else if (uid is UIDynamicTextField)
						RemoveTextField(uid as UIDynamicTextField);
					else if (uid is UIDynamicPopup)
					{
						// Workaround for VaM not cleaning its panels properly.
						UIDynamicPopup uidp = uid as UIDynamicPopup;
						RectTransform popupPanel = uidp.popup?.popupPanel;
						RemovePopup(uidp);
						if (popupPanel != null)
							UnityEngine.Object.Destroy(popupPanel.gameObject);
					}
                    else if (uid is UIDynamicV3Slider)
                    {
                        var v3Slider = uid as UIDynamicV3Slider;
                        leftUIElements.Remove(v3Slider.transform);
                        rightUIElements.Remove(v3Slider.spacer.transform);
                        Destroy(v3Slider.spacer.gameObject);
                        Destroy(v3Slider.gameObject);
                    }
                    else if (uid is UIDynamicTabBar)
                    {
                        var tabbar = uid as UIDynamicTabBar;
                        leftUIElements.Remove(tabbar.transform);
                        rightUIElements.Remove(tabbar.spacer.transform);
                        Destroy(tabbar.spacer.gameObject);
                        Destroy(tabbar.gameObject);
                    }
                    else
						RemoveSpacer(uid);
				}
			}

			UIElements.Clear();
		}
    }
    
    public class UIDynamicTabBar : UIDynamic
    {
        public int id;
        public int lastId;
        public List<Button> buttons = new List<Button>();
        public Action<int> SelectTab;
        public Action SelectLast;
        public UIDynamic spacer;

        public void ToggleButtons(bool val, int id = -1)
        {
            if (id == -1)
            {
                for (int i=0; i<buttons.Count; i++) buttons[i].gameObject.SetActive(val);
                return;
            }
            buttons[id].gameObject.SetActive(val);
        }
    }

    public class ChildrenChangedEventHandler : MonoBehaviour
    {
        private void OnTransformChildrenChanged()
        {
            foreach (Transform child in transform)
            {
                if (child.gameObject.GetComponent<ClothingEquippedEventHandler>()) continue;
                // child.GetComponent<DAZClothingItem>().uid.Print();
                child.gameObject.AddComponent<ClothingEquippedEventHandler>();
            }
        }
    }
    
    public class ClothingEquippedEventHandler : MonoBehaviour
    {
        private void OnEnable()
        {
            var dci = gameObject.GetComponent<DAZClothingItem>();
            HeelAdjust2.onClothingItemToggled(dci, true);
            // $"{dci.uid} equipped".Print();
        }

        private void OnDisable()
        {
            var dci = gameObject.GetComponent<DAZClothingItem>();
            HeelAdjust2.onClothingItemToggled(dci, false);
            // $"{gameObject.GetComponent<DAZClothingItem>().uid} unequipped".Print();
        }
    }

    public static class Extensions
    {
        public static void Print(this object s)
        {
            SuperController.LogMessage(s.ToString());
        }

        public static void Print(this Vector3 s)
        {
            SuperController.LogMessage(s.ToString());
        }
        
        public static void NullCheck(this object s)
        {
            SuperController.LogMessage((s == null).ToString());
        }

        public static Quaternion SetX(this Quaternion q, float val)
        {
            var v = q.eulerAngles;
            v.x = val;
            return Quaternion.Euler(v);
        }
        
        public static Atom GetAtom(this GameObject go)
        {
            Atom atom = go.GetComponent<Atom>();
            Transform parent = go.transform.parent;
            int i = 0;
            while (atom == null && parent != null && i < 20)
            {
                // atom = parent.GetComponent<Atom>();
                atom = parent.GetComponent<JSONStorable>()?.containingAtom;
                parent = parent.transform.parent;
                i++;
            }

            return atom;
        }
        
        public static void PrintChildren(this Transform t)
        {
            for (int i = 0; i < t.childCount; i++)
            {
                t.GetChild(i).name.Print();
            }
        }
		
        public static Transform[] GetChildren(this Transform t)
        {
            Transform[] children = new Transform[t.childCount];
            for (int i = 0; i < t.childCount; i++)
            {
                children[i] = t.GetChild(i);
            }
            return children;
        }

        public static bool Store(this JSONStorableParam jsonParam, JSONClass jc, bool forceStore = true)
        {
            bool needsStore = false;
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                if (forceStore || jParam.val != jParam.defaultVal)
                {
                    jc[jParam.name].AsBool = jParam.val;
                    needsStore = true;
                }
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                if (forceStore || jParam.val != jParam.defaultVal)
                {
                    jc[jParam.name].AsFloat = jParam.val;
                    needsStore = true;
                }
            }
            return needsStore;
        }

        public static void Load(this JSONStorableParam jsonParam, JSONClass jc)
        {
            if (jsonParam is JSONStorableBool)
            {
                JSONStorableBool jParam = (JSONStorableBool)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsBool;
            }
            else if (jsonParam is JSONStorableFloat)
            {
                JSONStorableFloat jParam = (JSONStorableFloat)jsonParam;
                if (jc.HasKey(jParam.name)) jParam.val = jc[jParam.name].AsFloat;
            }
        }
        
        public static JSONArray ToJA(this Vector3 v)
        {
            JSONArray ja = new JSONArray();
            for (int i = 0; i < 3; i++)
                ja[i].AsFloat = v[i];
            return ja;
        }
		
        public static Vector3 ToV3(this JSONArray ja)
        {
            Vector3 v = Vector3.zero;
            for (int i = 0; i < 3; i++)
                v[i] = ja[i].AsFloat;
            return v;
        }
        
        public static string GetPackagePath(this MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            int idx = filename.IndexOf(":/");
            if (idx >= 0) return filename.Substring(0, idx+2);
            return string.Empty;
        }
    }
}

